package CalculiX;

// This class describes an object that is able to build
// expressions of each kind. We refer to such an object as an
// "expression factory".

// When constructing abstract syntax trees, the parser relies on an
// expression factory instead of creating objects directly via "new".
// This allows the parser to be self-contained (that is, even if the
// various sub-classes of Expression have not been defined yet, the
// parser can be compiled). This is exploited by TD Interpretation.

class ExpressionFactory {

    Value buildValue (int i){return new VInteger(i);}

    Value buildValue (boolean b){return new VBoolean(b);}

  public Expression buildConstant (int i)
  {
    return new EConstant (i);
  }

  public Expression buildConstant (boolean b)
  {
    return new EConstant (b);
  }

  public Expression buildIntegerOperation (Expression left, Operator op, Expression right)
  {
    return new EIntegerOperation (left, op, right);
  }

  public Expression buildBooleanOperation (Expression left, Connective op, Expression right)
  {
    return new EBooleanOperation (left, op, right);
  }

  public Expression buildBooleanNegation (Expression e)
  {
    return new EBooleanOperation (e, Connective.XOR, new EConstant (true));
  }

  public Expression buildIntegerComparison (Expression left, Comparator op, Expression right)
  {
    return new EIntegerComparison (left, op, right);
  }

  public Expression buildIf (Expression condition, Expression then_branch, Expression else_branch)
  {
    return new EIf (condition, then_branch, else_branch);
  }

  public Expression buildLet (String x, Expression left, Expression right)
  {
    return new ELet (x, left, right);
  }

  public Expression buildVarRead (String x)
  {
    return new EVarRead (x);
  }

  public Expression buildSigma  (String x, Expression lo, Expression hi, Expression body)
  {
    return new ESigma  (x, lo, hi, body);
  }

}

class VInteger extends Value {

  // The integer constant.
  final int i;

  VInteger (int i)
  {
    this.i = i;
  }

  @Override  
  int asInteger ()
  {
    return i;
  }
  
  public String toString ()
  {
    return (new Integer (i)).toString();
  }

}

class VBoolean extends Value {

  // The Boolean constant.
  final boolean b;

  VBoolean (boolean b)
  {
    this.b = b;
  }

  @Override  
  boolean asBoolean ()
  {
    return b;
  }

  public String toString ()
  {
    return (new Boolean (b)).toString();
  }

}




class EConstant extends Expression {

  // The constant.
  final Value k;

  EConstant (int i)
  {
    this.k = new VInteger (i);
  }

  EConstant (boolean b)
  {
    this.k = new VBoolean (b);
  }
  
  Value eval (Environment env)
  {
    // A constant is its own value.
    return k;
  }

  public String toString(){
    return k.toString();
  }

}





class EIntegerOperation extends Expression {

  // The operator.
  final Operator op;
  // The operands.
  final Expression left, right;

  EIntegerOperation (Expression left, Operator op, Expression right)
  {
    this.left = left;
    this.op = op;
    this.right = right;
  }

  Value eval (Environment env) throws EvalException
  {
    // Evaluate the operands.
    Value v1 = left.eval(env);
    Value v2 = right.eval(env);
    // Make sure that they produce integer results.
    int i1 = v1.asInteger();
    int i2 = v2.asInteger();
    // Perform the appropriate operation.
    int i;
    switch (op) {
    case ADDITION:
      i = i1 + i2;
      break;
    case SUBTRACTION:
      i = i1 - i2;
      break;
    case MULTIPLICATION:
      i = i1 * i2;
      break;
    case DIVISION:
      i = i1 / i2;
      break;
    default:
      // This cannot happen.
      throw new AssertionError ();
    }
    // Package up the result as a value.
    return new VInteger (i);
    }

  public String toString(){
	  String s = "";
	  if (left instanceof EIntegerOperation){
		  Operator l = ((EIntegerOperation)left).op;
		  if ((l == Operator.ADDITION || l == Operator.SUBTRACTION) &&
		      (op == Operator.MULTIPLICATION || op == Operator.DIVISION)){
			  s += "(" + left.toString() + ")";
		  }
		  else s += left.toString();
	  }
	  else if (left instanceof EIf)	s += "(" + left.toString() + ")";
	  else s += left.toString();
	  
	  switch (op) {
	    case ADDITION:
	    	s += " + "; break;
	    case SUBTRACTION:
	    	s += " - "; break;
	    case MULTIPLICATION:
	    	s += " * "; break;
	    case DIVISION:
	    	s += " / "; break;
	    default:
	      // This cannot happen.
	      throw new AssertionError ();
	  }
			
	  
	  if (right instanceof EIntegerOperation){
		  Operator r = ((EIntegerOperation)right).op;
		  if ((r == Operator.ADDITION || r == Operator.SUBTRACTION) &&
		      (op == Operator.MULTIPLICATION || op == Operator.DIVISION || op == Operator.SUBTRACTION)){
			  s = s + "(" + right.toString() + ")";
		  }
		  else s += right.toString();
	  }
	  else s += right.toString(); 
			  
	return s;
	  
  }

}



class EBooleanOperation extends Expression {

  // The connective.
  final Connective op;
  // The operands.
  final Expression left, right;

  EBooleanOperation (Expression left, Connective op, Expression right)
  {
    this.left = left;
    this.op = op;
    this.right = right;
  }

  Value eval (Environment env) throws EvalException
  {
    // Evaluate the operands.
    Value v1 = left.eval(env);
    Value v2 = right.eval(env);
    // Make sure that they produce Boolean results.
    boolean b1 = v1.asBoolean();
    boolean b2 = v2.asBoolean();
    // Perform the appropriate operation.
    boolean b;
    switch (op) {
    case AND:
      b = b1 && b2;
      break;
    case OR:
      b = b1 || b2;
      break;
    case XOR:
      b = b1 ^ b2;
      break;
    default:
      // This cannot happen.
      throw new AssertionError ();
    }
    // Package up the result as a value.
    return new VBoolean (b);
  }

  public String toString(){
	  String s = "";
	  if (left instanceof EBooleanOperation){
		  if ((((EBooleanOperation)left).op == Connective.OR && op == Connective.AND) ||
				  ((EBooleanOperation)left).op != Connective.XOR && op == Connective.XOR)
   		     s += "(" + left.toString() + ")";
   		  else s += left.toString();
	  }
	  else if (left instanceof EIf)	s += "(" + left.toString() + ")";
	  else s += left.toString();
	  
	  switch (op) {
	    case OR: s += " || "; break;
	    case AND: s += " && "; break;
	    case XOR: s+= " ^ "; break;
	    default:
	      // This cannot happen.
	      throw new AssertionError ();
	    }
			  
	  if (right instanceof EBooleanOperation){
		  if ((((EBooleanOperation)right).op == Connective.OR && op == Connective.AND) ||
				  ((EBooleanOperation)right).op != Connective.XOR && op == Connective.XOR)
   		     s += "(" + right.toString() + ")";
   		  else s += right.toString();
	  }
	  else s += right.toString();		  
	  return s;
	  
  }

}
    

class EIntegerComparison extends Expression {

  // The comparator.
  final Comparator op;
  // The operands.
  final Expression left, right;

  EIntegerComparison (Expression left, Comparator op, Expression right)
  {
    this.left = left;
    this.op = op;
    this.right = right;
  }

  Value eval (Environment env) throws EvalException
  {
    // Evaluate the operands.
    Value v1 = left.eval(env);
    Value v2 = right.eval(env);
    // Make sure that they produce integer results.
    int i1 = v1.asInteger();
    int i2 = v2.asInteger();
    // Perform the appropriate operation.
    boolean b;
    switch (op) {
    case LESSTHAN:
      b = i1 < i2;
      break;
    case LESSTHANOREQUALTO:
      b = i1 <= i2;
      break;
    case EQUALS:
	b = (i1 == i2);
      break;
    default:
      // This cannot happen.
      throw new AssertionError ();
    }
    // Package up the result as a value.
    return new VBoolean (b);
    }

  public String toString(){
	  String s = "";
	  if (left instanceof EIntegerOperation){
		  s += left.toString();
	  }
	  else if (left instanceof EIf)	s += "(" + left.toString() + ")";
	  else s += left.toString();
	  
	  switch (op) {
	    case LESSTHAN: s += " < "; break;
	    case LESSTHANOREQUALTO: s += " <= "; break;
	    case EQUALS: s += " == "; break;
        default:
	      // This cannot happen.
	      throw new AssertionError ();
	    }
	  if (right instanceof EIntegerOperation){
		  s += "(" + right.toString() + ")";
	  }
	  else s += right.toString();
	  return s;
  }

}

class EIf extends Expression {

  // The condition.
  final Expression condition;
  // The branches.
  final Expression then_branch, else_branch;

  EIf (Expression condition, Expression then_branch, Expression else_branch)
  {
    this.condition = condition;
    this.then_branch = then_branch;
    this.else_branch = else_branch;
  }

  Value eval (Environment env) throws EvalException
  {
    // Evaluate the condition. We expect it to produce a Boolean value.
    // According to this condition, evaluate the appropriate branch.
    if (condition.eval(env).asBoolean())
      return then_branch.eval(env);
    else
      return else_branch.eval(env);
  }

  public String toString(){
	  String s = "if ";
    s += condition.toString();
	  s += " then ";
    s += then_branch.toString();
	  s += " else ";
    s += else_branch.toString();
	  return s;
  }

}




class EVarRead extends Expression {

  // The variable that is being read.
  final String x;

  EVarRead (String x)
  {
    this.x = x;
  }
  
  Value eval (Environment env) throws EvalException
  {
    // Query the environment to find the value associated with
    // this variable.
    return env.search(x);
  }

  public String toString(){
    return x;
  }

}

class ELet extends Expression {

  // The variable that is being declared.
  final String x;
  // The value with which this variable is initialized.
  final Expression left;
  // The expression that must be evaluated next.
  final Expression right;

  ELet (String x, Expression left, Expression right)
  {
    this.x = x;
    this.left = left;
    this.right = right;
  }

  Value eval (Environment env) throws EvalException
  {
    // Evaluate the left-hand expression in the environment env.
    // Extend the environment with a binding of x to this value.
    // Use the extended environment to evaluate the right-hand
    // expression.
    Value v1 = left.eval(env);
    return right.eval(env.extend(x, v1));
  }

  public String toString(){
    return "let "+x+"="+left+" in "+right;
  }

}



// This class implements an expression of the form
// Sigma_{x=lo}^{hi} {body}.

class ESigma extends Expression {

  // The variable that is used as an index.
  final String x;
  // The low and high bounds.
  final Expression lo, hi;
  // The body.
  final Expression body;

  ESigma (String x, Expression lo, Expression hi, Expression body)
  {
    this.x = x;
    this.lo = lo;
    this.hi = hi;
    this.body = body;
  }

  Value eval (Environment env) throws EvalException
  {
    int lo_value = lo.eval(env).asInteger();
    int hi_value = hi.eval(env).asInteger();
    int sum = 0;
    for (int v = lo_value; v <= hi_value; v++)
      sum += body.eval(
               env.extend(x, new VInteger (v))
             ).asInteger();
    return new VInteger (sum);
  }

}

